# ******************************************************************************
#  Copyright (c) 2021. Kneron Inc. All rights reserved.                        *
# ******************************************************************************

import numpy as np

'''
import ktc
EXEC_OPT = 1
'''
from utils.ExampleHelper import get_device_usb_speed_by_port_id
import kp
EXEC_OPT = 2


MODEL_PATH = '../../res/models/KL720/'
MODEL_FILE = MODEL_PATH + 'modInf/models_720.nef'

DATA_PATH = '../../res/images/'
DATA_FILE = DATA_PATH + 'modInf/tri_filtData_151.npy'

BOARD_VID = 0x3231
BOARD_PID = 0x0720


def get_product_name(product_id: int) -> str:
    if product_id == kp.ProductId.KP_DEVICE_KL520.value:
        return '(KL520)'
    elif product_id == kp.ProductId.KP_DEVICE_KL720.value:
        return '(KL720)'
    elif product_id == 0x200:
        return '(KL720) (Please update firmware by Kneron DFUT)'
    else:
        return '(Unknown Device)'


def ktc_Inference():
    frame_data = np.load(DATA_FILE)
    print( 'fr_shape', frame_data.shape, 'type', type(frame_data[0][0][0][0]) )

    # check nef radix from quantization data
    radix = ktc.get_radix(frame_data)
    print( 'radix', radix )

    trigger_count = []
    pred_max = 0

    for idx in range( len(frame_data) ):
        input_data = [ frame_data[idx] ]
        '''
        '''
        print( np.array(input_data).shape, end='\t' )
        print( '%f %f %f %f' % (
                    input_data[0][0][0][0],
                    input_data[0][0][1][0],
                    input_data[0][0][2][0],
                    input_data[0][0][3][0]
                    ), end='\t'
                )

        # 5.2 E2E Simulator Check (Hardware)
        nef_predict = ktc.kneron_inference(
                input_data,
                nef_file = MODEL_FILE,
                platform = 720,
                radix = radix
                )[0][0][0]

        if nef_predict[0][1] > 0.5:
            print( idx, 'TRI\t', nef_predict[0][0], nef_predict[0][1] )
        else:
            print( idx, '\t', nef_predict[0][0], nef_predict[0][1] )


def KL720_Inference():
    usb_port_id = 0

    print('scanning kneron devices ...')
    device_list = kp.core.scan_devices()

    print('number of Kneron devices found: {}'.format(device_list.device_descriptor_number))

    if device_list.device_descriptor_number == 0:
        exit(0)

    print('listing devices infomation as follows:')
    for idx, device_descriptor in enumerate(device_list.device_descriptor_list):
        print()
        print('[{}] USB scan index: \'{}\''.format(idx, idx))
        print('[{}] USB port ID: \'{}\''.format(idx, device_descriptor.usb_port_id))
        print('[{}] Product ID: \'0x{:X} {}\''.format(idx,
                                                      device_descriptor.product_id,
                                                      get_product_name(product_id=device_descriptor.product_id)))

        print('[{}] USB link speed: \'{}\''.format(idx, device_descriptor.link_speed))
        print('[{}] USB port path: \'{}\''.format(idx, device_descriptor.usb_port_path))
        print('[{}] KN number: \'0x{:X}\''.format(idx, device_descriptor.kn_number))
        print('[{}] Connectable: \'{}\''.format(idx, device_descriptor.is_connectable))
        print('[{}] Firmware: \'{}\''.format(idx, device_descriptor.firmware))

        if device_descriptor.vendor_id == BOARD_VID and device_descriptor.product_id == BOARD_PID:
            usb_port_id = device_descriptor.usb_port_id
    print()

    """
    check device USB speed (Recommend run KL720 at super speed)
    """
    try:
        if kp.UsbSpeed.KP_USB_SPEED_SUPER != get_device_usb_speed_by_port_id(usb_port_id=usb_port_id):
            print('\033[91m' + '[Warning] Device is not run at super speed.' + '\033[0m')
    except Exception as exception:
        print('Error: check device USB speed fail, port ID = \'{}\', error msg: [{}]'.format(usb_port_id,
                                                                                             str(exception)))
        exit(0)

    """
    connect the device
    """
    try:
        print('[Connect Device]', 'w/ usb_port_id', usb_port_id)
        device_group = kp.core.connect_devices(usb_port_ids=[usb_port_id])
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: connect device fail, port ID = \'{}\', error msg: [{}]'.format(usb_port_id,
                                                                                     str(exception)))
        exit(0)

    """
    setting timeout of the usb communication with the device
    """
    print('[Set Device Timeout]')
    kp.core.set_timeout(device_group=device_group, milliseconds=5000)
    print(' - Success')

    """
    load model from device flash
    """
    try:
        '''
        print('[Load Model from Flash]')
        model_nef_descriptor = kp.core.load_model_from_flash(device_group=device_group)
        '''
        print('[Load Model from File]')
        model_nef_descriptor = kp.core.load_model_from_file(device_group=device_group,
                                                            file_path=MODEL_FILE)
        print(model_nef_descriptor)
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: load model from device flash failed, error = \'{}\''.format(str(exception)))
        exit(0)

    """
    prepare the image
    """
    print('[Read Image]')
    frame_data = np.load(DATA_FILE)
    #frame_data = np.einsum( 'ijkl->ikjl', frame_data )
    frame_num = len(frame_data)
    frame_size = len( bytes(frame_data[0].reshape(-1)) )
    print( 'fr_shape', frame_data.shape, 'fr_size', frame_size, 'type', type(frame_data[0][0][0][0]) )
    print(' - Success')

    """
    prepare app generic inference config
    """
    generic_raw_image_header = kp.GenericRawBypassPreProcImageHeader(
        model_id=model_nef_descriptor.models[0].id,
        image_buffer_size=frame_size,
        inference_number=0
    )

    """
    starting inference work
    """
    print('[Starting Inference Work]')
    print(' - Starting inference loop {} times'.format(frame_num))
    print(' - ', end='')
    inference_results = []
    for i in range(frame_num):
        generic_raw_image_header.inference_number = i

        img_buf = bytes( frame_data[i].reshape(-1) )
        '''
        img_buf = np.zeros( (27, 29, 3), dtype=np.uint32 )
        img_buf = bytes( img_buf.reshape(-1) )
        '''

        try:
            kp.inference.generic_raw_inference_bypass_pre_proc_send(
                device_group=device_group,
                generic_raw_image_header=generic_raw_image_header,
                image_buffer=img_buf)

            generic_raw_bypass_pre_proc_result = kp.inference.generic_raw_inference_bypass_pre_proc_receive(
                device_group=device_group,
                generic_raw_image_header=generic_raw_image_header,
                model_nef_descriptor=model_nef_descriptor)
            inference_results.append(generic_raw_bypass_pre_proc_result)
        except kp.ApiKPException as exception:
            print(' - Error: inference failed, error = {}'.format(exception))
            exit(0)

        print('.', end='', flush=True)
    print()

    """
    retrieve inference node output 
    """
    print( '[Retrieve Inference Node Output ]', '#{}'.format(generic_raw_bypass_pre_proc_result.header.num_output_node) )
    inf_node_output_list = []
    for inf_idx in range( len(inference_results) ):
        inference_float_node_output = kp.inference.generic_inference_retrieve_float_node(
            node_idx=0,
            generic_raw_result=inference_results[inf_idx],
            channels_ordering=kp.ChannelOrdering.KP_CHANNEL_ORDERING_CHW)
        inf_node_output_list.append(inference_float_node_output)

    print(' - Success')

    print('\r\n[Result]')
    '''
    print(inf_node_output_list)
    '''
    for out_idx in range( len(inf_node_output_list) ):
        '''
        '''
        print( '%f %f %f %f' % (
                    frame_data[out_idx][0][0][0],
                    frame_data[out_idx][0][1][0],
                    frame_data[out_idx][0][2][0],
                    frame_data[out_idx][0][3][0]
                    ),
                end='\t'
                )

        out_data = inf_node_output_list[out_idx].ndarray
        #print( out_data.shape, out_data.dtype )
        # flat
        out_data = out_data.reshape(-1)
        if out_data[1] > 0.5:
            print( '{}\t{} [ {} ]\tTRI'.format( out_idx, out_data[0], out_data[1] ) )
        else:
            print( '{}\t{}   {}'.format( out_idx, out_data[0], out_data[1] ) )

    kp.core.disconnect_devices( device_group=device_group )


if __name__ == '__main__':
    if 1 == EXEC_OPT:
        ktc_Inference()
    elif 2 == EXEC_OPT:
        KL720_Inference()

